@@ -24,6 +24,7 @@
 #include <libedata-book/e-data-book.h>
 #include <libedata-book/e-data-book-view.h>
 #include <dbus/dbus-glib.h>
+#include <gnome-keyring.h>
 
 #define COUCHDB_REVISION_PROP                "X-COUCHDB-REVISION"
 #define COUCHDB_UUID_PROP                    "X-COUCHDB-UUID"
@@ -299,6 +300,66 @@ vcard_from_couch_document (CouchDBDocument *document)
 		couchdb_struct_field_unref (url);
 	}
 
+	/* parse IM addresses */
+	list = couchdb_document_contact_get_im_addresses (document);
+	while (list != NULL) {
+		const char *address_str, *description_str, *protocol_str, *uuid_str;
+		EVCardAttribute *attr = NULL;
+		CouchDBStructField *im = (CouchDBStructField *) list->data;
+
+		address_str = couchdb_document_contact_im_get_address (im);
+		description_str = couchdb_document_contact_im_get_description (im);
+		protocol_str = couchdb_document_contact_im_get_protocol (im);
+		uuid_str = couchdb_struct_field_get_uuid (im);
+
+		if (protocol_str != NULL) {
+			if (g_strcmp0 (protocol_str, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_AIM) == 0)
+				attr = e_vcard_attribute_new (NULL, e_contact_vcard_attribute (E_CONTACT_IM_AIM));
+			else if (g_strcmp0 (protocol_str, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_GADU_GADU) == 0)
+				attr = e_vcard_attribute_new (NULL, e_contact_vcard_attribute (E_CONTACT_IM_GADUGADU));
+			else if (g_strcmp0 (protocol_str, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_GROUPWISE) == 0)
+				attr = e_vcard_attribute_new (NULL, e_contact_vcard_attribute (E_CONTACT_IM_GROUPWISE));
+			else if (g_strcmp0 (protocol_str, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_ICQ) == 0)
+				attr = e_vcard_attribute_new (NULL, e_contact_vcard_attribute (E_CONTACT_IM_ICQ));
+			else if (g_strcmp0 (protocol_str, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_JABBER) == 0)
+				attr = e_vcard_attribute_new (NULL, e_contact_vcard_attribute (E_CONTACT_IM_JABBER));
+			else if (g_strcmp0 (protocol_str, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_MSN) == 0)
+				attr = e_vcard_attribute_new (NULL, e_contact_vcard_attribute (E_CONTACT_IM_MSN));
+			else if (g_strcmp0 (protocol_str, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_SKYPE) == 0)
+				attr = e_vcard_attribute_new (NULL, e_contact_vcard_attribute (E_CONTACT_IM_SKYPE));
+			else if (g_strcmp0 (protocol_str, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_YAHOO) == 0)
+				attr = e_vcard_attribute_new (NULL, e_contact_vcard_attribute (E_CONTACT_IM_YAHOO));
+
+			if (attr != NULL) {
+				if (description_str) {
+					if (!g_ascii_strcasecmp (description_str, "home")) {
+						e_vcard_attribute_add_param_with_value (attr,
+											e_vcard_attribute_param_new (EVC_TYPE),
+											"HOME");
+					} else if (!g_ascii_strcasecmp (description_str, "work")) {
+						e_vcard_attribute_add_param_with_value (attr,
+											e_vcard_attribute_param_new (EVC_TYPE),
+											"WORK");
+					}
+				}
+
+				if (uuid_str != NULL) {
+					EVCardAttributeParam *param;
+
+					param = e_vcard_attribute_param_new (COUCHDB_UUID_PROP);
+					e_vcard_attribute_add_param_with_value (attr, param, uuid_str);
+				}
+
+				e_vcard_attribute_add_value (attr, address_str);
+				e_vcard_add_attribute (E_VCARD (contact), attr);
+			}
+		}
+
+		/* remove addresses from list */
+		list = g_slist_remove (list, im);
+		couchdb_struct_field_unref (im);
+	}
+
 	/* birth date */
 	str = (char *) couchdb_document_contact_get_birth_date (document);
 	if (str) {
@@ -406,7 +467,7 @@ contact_phone_to_struct_field (EVCardAttribute *attr)
 		GList *v;
 		EVCardAttributeParam *p = pl->data;
 
-		if (!g_strcmp0 (EVC_TYPE, e_vcard_attribute_param_get_name (p)) != 0) {
+		if (g_strcmp0 (EVC_TYPE, e_vcard_attribute_param_get_name (p)) == 0) {
 			v = e_vcard_attribute_param_get_values (p);
 			while (v && v->data) {
 				if (g_ascii_strcasecmp ((const gchar *) v->data, "HOME") == 0)
@@ -473,7 +534,7 @@ contact_url_to_struct_field (EVCardAttribute *attr, const gchar *description)
 		GList *v;
 		EVCardAttributeParam *p = pl->data;
 
-		if (!g_strcmp0 (COUCHDB_UUID_PROP, e_vcard_attribute_param_get_name (p)) != 0) {
+		if (g_strcmp0 (COUCHDB_UUID_PROP, e_vcard_attribute_param_get_name (p)) == 0) {
 			v = e_vcard_attribute_param_get_values (p);
 			if (v && v->data)
 				uuid = (const gchar *) v->data;
@@ -483,6 +544,44 @@ contact_url_to_struct_field (EVCardAttribute *attr, const gchar *description)
 	return couchdb_document_contact_url_new (uuid, address, description);
 }
 
+static CouchDBStructField *
+contact_im_to_struct_field (EVCardAttribute *attr, const gchar *protocol)
+{
+	const gchar *address, *description = NULL, *uuid = NULL;
+	GList *params, *pl;
+
+	address = e_vcard_attribute_get_value (attr);
+	if (!address)
+		return NULL;
+
+	params = e_vcard_attribute_get_params (attr);
+	if (!params)
+		return couchdb_document_contact_im_new (NULL, address, "other", protocol);
+
+	for (pl = params; pl != NULL; pl = pl->next) {
+		GList *v;
+		EVCardAttributeParam *p = pl->data;
+
+		if (g_strcmp0 (COUCHDB_UUID_PROP, e_vcard_attribute_param_get_name (p)) == 0) {
+			v = e_vcard_attribute_param_get_values (p);
+			if (v && v->data)
+				uuid = (const gchar *) v->data;
+		} else if (g_strcmp0 (EVC_TYPE, e_vcard_attribute_param_get_name (p)) == 0) {
+			v = e_vcard_attribute_param_get_values (p);
+			if (v && v->data) {
+			        if (g_strcmp0 ("HOME", (const gchar *) v->data) == 0)
+					description = "home";
+				else if (g_strcmp0 ("WORK", (const gchar *) v->data) == 0)
+					description = "work";
+				else
+					description = "other";
+			}
+		}
+	}
+
+	return couchdb_document_contact_im_new (uuid, address, description, protocol);
+}
+
 static CouchDBDocument *
 couch_document_from_contact (EBookBackendCouchDB *couchdb_backend, EContact *contact)
 {
@@ -638,6 +737,49 @@ couch_document_from_contact (EBookBackendCouchDB *couchdb_backend, EContact *con
 		g_slist_free (list);
 	}
 
+	/* IM addresses */
+	list = NULL;
+	attr_list = e_vcard_get_attributes (E_VCARD (contact));
+	for (al = attr_list; al != NULL; al = al->next) {
+		CouchDBStructField *sf = NULL;
+		EVCardAttribute *attr = (EVCardAttribute *) al->data;
+
+		if (g_strcmp0 (e_vcard_attribute_get_name (attr),
+			       e_contact_vcard_attribute (E_CONTACT_IM_AIM)) == 0)
+			sf = contact_im_to_struct_field (attr, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_AIM);
+		else if (g_strcmp0 (e_vcard_attribute_get_name (attr),
+			       e_contact_vcard_attribute (E_CONTACT_IM_GADUGADU)) == 0)
+			sf = contact_im_to_struct_field (attr, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_GADU_GADU);
+		else if (g_strcmp0 (e_vcard_attribute_get_name (attr),
+			       e_contact_vcard_attribute (E_CONTACT_IM_GROUPWISE)) == 0)
+			sf = contact_im_to_struct_field (attr, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_GROUPWISE);
+		else if (g_strcmp0 (e_vcard_attribute_get_name (attr),
+			       e_contact_vcard_attribute (E_CONTACT_IM_ICQ)) == 0)
+			sf = contact_im_to_struct_field (attr, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_ICQ);
+		else if (g_strcmp0 (e_vcard_attribute_get_name (attr),
+			       e_contact_vcard_attribute (E_CONTACT_IM_JABBER)) == 0)
+			sf = contact_im_to_struct_field (attr, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_JABBER);
+		else if (g_strcmp0 (e_vcard_attribute_get_name (attr),
+			       e_contact_vcard_attribute (E_CONTACT_IM_MSN)) == 0)
+			sf = contact_im_to_struct_field (attr, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_MSN);
+		else if (g_strcmp0 (e_vcard_attribute_get_name (attr),
+			       e_contact_vcard_attribute (E_CONTACT_IM_SKYPE)) == 0)
+			sf = contact_im_to_struct_field (attr, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_SKYPE);
+		else if (g_strcmp0 (e_vcard_attribute_get_name (attr),
+			       e_contact_vcard_attribute (E_CONTACT_IM_YAHOO)) == 0)
+			sf = contact_im_to_struct_field (attr, COUCHDB_DOCUMENT_CONTACT_IM_PROTOCOL_YAHOO);
+
+		if (sf != NULL)
+			list = g_slist_append (list, sf);
+	}
+
+	if (list != NULL) {
+		couchdb_document_contact_set_im_addresses (document, list);
+
+		g_slist_foreach (list, (GFunc) couchdb_struct_field_unref, NULL);
+		g_slist_free (list);
+	}
+
 	/* birth date */
 	dt = (EContactDate *) e_contact_get_const (contact, E_CONTACT_BIRTH_DATE);
 	if (dt) {
@@ -671,6 +813,40 @@ couch_document_from_contact (EBookBackendCouchDB *couchdb_backend, EContact *con
 	return document;
 }
 
+static void
+document_updated_cb (CouchDB *couchdb, const char *dbname, CouchDBDocument *document, gpointer user_data)
+{
+	char *vcard;
+	EContact *contact;
+	EBookBackendCouchDB *couchdb_backend = E_BOOK_BACKEND_COUCHDB (user_data);
+
+	if (g_strcmp0 (dbname, couchdb_backend->dbname) != 0)
+		return;
+
+	vcard = vcard_from_couch_document (document);
+	if (!vcard)
+		return;
+
+	contact = e_contact_new_from_vcard (vcard);
+	if (contact != NULL) {
+		e_book_backend_notify_update (E_BOOK_BACKEND (couchdb_backend), contact);
+		g_object_unref (G_OBJECT (contact));
+	}
+
+	g_free (vcard);
+}
+
+static void
+document_deleted_cb (CouchDB *couchdb, const char *dbname, const char *docid, gpointer user_data)
+{
+	EBookBackendCouchDB *couchdb_backend = E_BOOK_BACKEND_COUCHDB (user_data);
+
+	if (g_strcmp0 (dbname, couchdb_backend->dbname) != 0)
+		return;
+
+	e_book_backend_notify_remove (E_BOOK_BACKEND (couchdb_backend), docid);
+}
+
 static GNOME_Evolution_Addressbook_CallStatus
 e_book_backend_couchdb_load_source (EBookBackend *backend,
 				    ESource *source,
@@ -680,6 +856,8 @@ e_book_backend_couchdb_load_source (EBookBackend *backend,
 	const gchar *property;
 	CouchDBDatabaseInfo *db_info;
 	GError *error = NULL;
+	gboolean enable_oauth = FALSE;
+	char *oauth_c_key = NULL, *oauth_c_secret = NULL, *oauth_t_key = NULL, *oauth_t_secret = NULL;
 	EBookBackendCouchDB *couchdb_backend = E_BOOK_BACKEND_COUCHDB (backend);
 
 	g_return_val_if_fail (E_IS_BOOK_BACKEND_COUCHDB (couchdb_backend), GNOME_Evolution_Addressbook_OtherError);
@@ -722,9 +900,43 @@ e_book_backend_couchdb_load_source (EBookBackend *backend,
 		g_object_unref (G_OBJECT (proxy));
 		dbus_g_connection_unref (bus);
 
-		if (success)
+		if (success) {
+			GnomeKeyringAttributeList *attrs;
+			GnomeKeyringResult result;
+			GList *items_found;
+
 			uri = g_strdup_printf ("http://127.0.0.1:%d", port);
-		else {
+
+			/* Get OAuth tokens from GnomeKeyring */
+			attrs = gnome_keyring_attribute_list_new();
+			gnome_keyring_attribute_list_append_string (attrs, "desktopcouch", "oauth");
+
+			result = gnome_keyring_find_items_sync (GNOME_KEYRING_ITEM_GENERIC_SECRET,
+								attrs, &items_found);
+			if (result == GNOME_KEYRING_RESULT_OK && items_found != NULL) {
+				gchar **items;
+				GnomeKeyringFound *first_item = (GnomeKeyringFound *) items_found->data;
+
+				items = g_strsplit (first_item->secret, ":", 4);
+				if (items) {
+					oauth_c_key = g_strdup (items[0]);
+					oauth_c_secret = g_strdup (items[1]);
+					oauth_t_key = g_strdup (items[2]);
+					oauth_t_secret = g_strdup (items[3]);
+					g_strfreev (items);
+
+					enable_oauth = TRUE;
+				}
+
+				gnome_keyring_found_list_free (items_found);
+			} else {
+				g_warning ("Could not get OAuth tokens from keyring: %s",
+					   gnome_keyring_result_to_message (result));
+			}
+
+			/* Free memory */
+			gnome_keyring_attribute_list_free (attrs);
+		} else {
 			g_warning ("Couldn't get port for desktopcouch: %s", error->message);
 			g_error_free (error);
 
@@ -741,6 +953,19 @@ e_book_backend_couchdb_load_source (EBookBackend *backend,
 		return GNOME_Evolution_Addressbook_OtherError;
 	}
 
+	if (enable_oauth) {
+		couchdb_enable_oauth (couchdb_backend->couchdb,
+				      oauth_c_key,
+				      oauth_c_secret,
+				      oauth_t_key,
+				      oauth_t_secret);
+
+		g_free (oauth_c_key);
+		g_free (oauth_c_secret);
+		g_free (oauth_t_key);
+		g_free (oauth_t_secret);
+	}
+
 	g_free (uri);
 
 	/* check if only_if_exists */
@@ -773,6 +998,15 @@ e_book_backend_couchdb_load_source (EBookBackend *backend,
 	e_book_backend_set_is_loaded (backend, TRUE);
 	e_book_backend_set_is_writable (backend, TRUE);
 
+	/* Listen for changes on database */
+	g_signal_connect (G_OBJECT (couchdb_backend->couchdb), "document_created",
+			  G_CALLBACK (document_updated_cb), couchdb_backend);
+	g_signal_connect (G_OBJECT (couchdb_backend->couchdb), "document_updated",
+			  G_CALLBACK (document_updated_cb), couchdb_backend);
+	g_signal_connect (G_OBJECT (couchdb_backend->couchdb), "document_deleted",
+			  G_CALLBACK (document_deleted_cb), couchdb_backend);
+	couchdb_listen_for_changes (couchdb_backend->couchdb, couchdb_backend->dbname);
+
 	return GNOME_Evolution_Addressbook_Success;
 }
 
@@ -1115,6 +1349,16 @@ e_book_backend_couchdb_get_supported_fields (EBookBackend *backend,
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_ADDRESS_WORK)));
 	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_ADDRESS_OTHER)));
 
+	/* IM addresses */
+	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_IM_AIM)));
+	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_IM_GADUGADU)));
+	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_IM_GROUPWISE)));
+	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_IM_ICQ)));
+	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_IM_JABBER)));
+	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_IM_MSN)));
+	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_IM_SKYPE)));
+	fields = g_list_append (fields, g_strdup (e_contact_field_name (E_CONTACT_IM_YAHOO)));
+
 	e_data_book_respond_get_supported_fields (book, opid,
 						  GNOME_Evolution_Addressbook_Success, fields);
 
